#!/usr/bin/env python
"""Feed this script the path of a Python module and it will print out
any warnings from pylint, pyflakes, and pep8.py that haven't been
explicitly silenced. The default output format is compatible with
Emacs' flymake mode.

This is derived from the anonymous flymake wrapper at
http://www.emacswiki.org/emacs/PythonMode
I'd credit the author, if I knew who they were.
"""
# Copyright (c) 2002-present, Damn Simple Solutions Ltd.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#     * Neither the name of the Damn Simple Solutions Ltd. nor the names
#       of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
from optparse import OptionParser
from subprocess import Popen, PIPE
import os


FLYMAKE_OUTPUT_FORMAT = (
    "%(level)s %(error_type)s%(error_number)s:"
    "%(description)s at %(filename)s line %(line_number)s.")


def _find_root_dir():
    """Find top level of project."""
    source = os.environ.get('SRCDIR', os.getcwd())
    root = source
    while os.sep in root and not os.path.exists(
        os.path.join(root, '.bzr')):
        root = os.sep.join(root.split(os.sep)[:-1])
    if os.sep in root:
        return root
    return source


SRCDIR = _find_root_dir() or os.getcwd()


def find_pylintrc():
    """Return location of pylintrc."""
    path = os.path.join(_find_root_dir(), 'pylintrc')
    if os.path.exists(path):
        return path
    path = os.path.join(_find_root_dir(), 'utilities', 'pylintrc')
    if os.path.exists(path):
        return path
    path = os.path.join(_find_root_dir(), 'dev', 'pylintrc')
    if os.path.exists(path):
        return path

    return '/usr/share/ubuntuone-dev-tools/pylintrc'


class LintRunner(object):
    """Lint runner."""
    sane_default_ignore_codes = set([])
    command = None
    output_matcher = None
    output_format = FLYMAKE_OUTPUT_FORMAT
    env = None

    def __init__(self, ignore_codes=(), use_sane_defaults=True):
        self.ignore_codes = set(ignore_codes)
        self.use_sane_defaults = use_sane_defaults

    @property
    def operative_ignore_codes(self):
        """Ignore codes."""
        if self.use_sane_defaults:
            return self.ignore_codes.union(self.sane_default_ignore_codes)
        else:
            return self.ignore_codes

    @property
    def run_flags(self):
        """Run flags."""
        return []

    def fixup_data(self, line, data):   # pylint: disable=W0613
        """Fix data."""
        if self.command:
            data['error_number'] = '%s %s' % (
                data.get('error_number', '?'), self.command)
        return data

    def process_line(self, line):
        """Process a single line."""
        m = self.output_matcher.match(line)
        if m:
            fixed_data = dict.fromkeys(
                ('level', 'error_type', 'error_number', 'description',
                 'filename', 'line_number'), '')
            fixed_data.update(self.fixup_data(line, m.groupdict()))
            self._handle_output(line, fixed_data)

    def _handle_output(self, line, fixed_data):  # pylint: disable=W0613
        """Handle the output."""
        print self.output_format % fixed_data

    def run(self, filenames):
        """Run the lint command."""
        args = [self.command]
        args.extend(self.run_flags)
        args.extend(filenames)
        process = Popen(args, stdout=PIPE, stderr=PIPE, env=self.env)
        for line in process.stdout:
            self.process_line(line)
        for line in process.stderr:
            self.process_line(line)


class PylintRunner(LintRunner):
    """Pylint runner."""

    output_matcher = re.compile(
        r'(?P<filename>[^:]+):'
        r'(?P<line_number>\d+):'
        r'\s*\[(?P<error_type>[WECR])(?P<error_number>[^,\]]+)[^\]]*\]'
        r'\s*(?P<description>.*)$')

    command = 'pylint'
    sane_default_ignore_codes = set([])

    env = {'PYTHONPATH': SRCDIR}

    def fixup_data(self, line, data):
        data = LintRunner.fixup_data(self, line, data)
        if data['error_type'].startswith('E'):
            data['level'] = 'ERROR'
        else:
            data['level'] = 'WARNING'
        return data

    @property
    def run_flags(self):
        return ('--output-format', 'parseable', '--include-ids', 'y',
                '--reports', 'n',
                '--rcfile', find_pylintrc())


class Flake8Runner(LintRunner):
    """Flake8 runner."""

    command = 'flake8'
    output_matcher = re.compile(
        r'(?P<filename>[^:]+):'
        r'(?P<line_number>[^:]+):'
        r'([^:]+:)?'
        r'\s*((?P<error_type>[WECR])(?P<error_number>[^\s]+))?'
        r'\s*(?P<description>.+)$')

    def fixup_data(self, line, data):
        data = LintRunner.fixup_data(self, line, data)
        if data['error_type']:
            if data['error_type'].startswith('E'):
                data['level'] = 'ERROR'
            else:
                data['level'] = 'WARNING'
        else:
            data['level'] = 'WARNING'
        return data

    def _handle_output(self, line, fixed_data):
        if 'redefinition of unused' in fixed_data['description']:
            return
        elif 'redefinition of function' in fixed_data['description']:
            return
        else:
            print self.output_format % fixed_data


class Pep8Runner(LintRunner):
    """Pep8 runner."""

    command = 'pep8'
    sane_default_ignore_codes = set([
        'E501',    # line to long handled by pyflakes
        'E113'])   # unexpected indentation i.e. bad syntax - is handled
                   # by pyflakes
    output_matcher = re.compile(
        r'(?P<filename>[^:]+):'
        r'(?P<line_number>[^:]+):'
        r'[^:]+:'
        r' (?P<error_number>\w+) '
        r'(?P<description>.+)$')

    def fixup_data(self, line, data):
        data = LintRunner.fixup_data(self, line, data)
        if ('W' in data['error_number']
            or data['error_number'][1] in '2345'):
            data['level'] = 'WARNING'
        else:
            data['level'] = 'ERROR'

        return data

    @property
    def run_flags(self):
        return (
            '--repeat', '--filename', '*py', '--ignore',
            ','.join(self.operative_ignore_codes))


class PyflakesRunner(LintRunner):
    """Pyflakes runner."""

    command = 'pyflakes'
    output_matcher = re.compile(
        r'(?P<filename>[^:]+):'
        r'(?P<line_number>[^:]+)\s*:'
        r'(?P<description>.+)$')

    def fixup_data(self, line, data):
        data = LintRunner.fixup_data(self, line, data)
        data['level'] = 'ERROR'
        return data

    def _handle_output(self, line, fixed_data):
        if 'redefinition of unused' in fixed_data['description']:
            return
        elif 'redefinition of function' in fixed_data['description']:
            return
        else:
            print self.output_format % fixed_data


CHECKERS = {
    "pylint": PylintRunner,
    "flake8": Flake8Runner}


def main():
    """Main"""
    usage = "usage: %prog [options] [PY_MODULE_OR_PACKAGE]..."
    parser = OptionParser(usage=usage)
    parser.add_option(
        "-c", "--comint_format",
        action="store_true",
        default=False,
        help=("Emacs comint compatible output format."
              " Flymake compatible is the default."))
    parser.add_option(
        "-i", "--ignore_codes",
        default="",
        help="comma separated list of error codes to ignore")
    parser.add_option(
        "-p", "--checker_progs",
        default="flake8,pylint",
        help=("comma separated list of the checker"
              " programs to run. default=flake8,pylint"))
    options, filenames = parser.parse_args()
    #

    selected_checkers = [
        CHECKERS[c]
        for c in options.checker_progs.strip().split(",")]

    ignore_codes = (
        options.ignore_codes.split(",") if options.ignore_codes else [])

    for runner_class in selected_checkers:
        runner = runner_class(ignore_codes=ignore_codes)
        runner.run(filenames)

if __name__ == '__main__':
    main()
