#!/usr/bin/python
"""Mail sorter for notmuch."""

# Warning:
# Be careful about using Query.count_messages(), it's technically an estimate
# and is not guarranteed to be correct

import logging
import time
import subprocess
from subprocess import PIPE
import notmuch
import re


def tag_message(msg, *tags):
    msg.freeze()
    for tag in tags:
        if tag[0] == '+':
            msg.add_tag(tag[1:])
        elif tag[0] == '-':
            msg.remove_tag(tag[1:])
        else:
            msg.add_tag(tag)
    msg.thaw()


def tag_search(db, search, *tags):
    q = notmuch.Query(db, search)
    count = 0
    for msg in q.search_messages():
        count += 1
        tag_message(msg, *tags)
    if count > 0:
        logging.debug(
            'Tagging %d messages with (%s)', count, ' '.join(tags))


def sort_junk(messages):
    spam_re = re.compile('X-Bogosity:\s*Spam')
    bf = subprocess.Popen(['bogofilter', '-bv'], stdin=PIPE, stdout=PIPE)
    for msg in messages:
        bf.stdin.write(msg.get_filename() + '\n')
        l = bf.stdout.readline()
        if re.search(spam_re, l):
            logging.debug('Message %s marked as junk', msg.get_message_id())
            msg.add_tag('junk')
            msg.remove_tag('new')
    bf.stdin.close()

logging.basicConfig(level=logging.DEBUG)

TAGS = []
start_time = time.time()


def tag(search, *tags):
    TAGS.append((search, tags))

tag('to:thisfred@gmail.com', 'gmail')
tag('to:eric.casteleijn@canonical.com', 'canonical')
tag('from:thisfred@gmail.com', 'sent', 'gmail')
tag('from:eric.casteleijn@canonical.com', 'sent', 'canonical')
tag('from:"eric casteleijn" and not sent', 'sent')

db = notmuch.Database(mode=notmuch.Database.MODE.READ_WRITE)

# Freeze new messages
q_new_msgs = notmuch.Query(db, 'tag:new')
n_msgs = 0

new_messages = q_new_msgs.search_messages()
for msg in new_messages:
    msg.freeze()
    n_msgs += 1

# Take care of basics
tag_search(db, 'tag:new', '+unread')


# Run through Bogofilter
sort_junk(q_new_msgs.search_messages())

# Tag things
for search, tags in TAGS:
    tag_search(  # pylint: disable=W0142
        db, '( %s ) and tag:new' % search, *tags)

# Ignore and watch things I sent
tag_search(db, 'tag:new and tag:sent', '-new')
tag_search(db, 'tag:new and folder:gmail/[Gmail].Forums', '+lists')
tag_search(db, 'tag:new and folder:gmail/[Gmail].Bulk', '+bulk')
tag_search(
    db, 'tag:new and subject:"[Bug "', '+lists', '+bugs',
    '+launchpad')
tag_search(
    db, 'tag:new and subject:"[Merge] lp:~"', '+lists', '+mergeproposals',
    '+launchpad')
tag_search(
    db, 'tag:new and subject:5kamp', '+lists', '+5kamp')
tag_search(
    db, 'tag:new and subject:"[Ubunet-discuss]"', '+ubuntuone',
    '+lists', '+ubunet-discuss')
tag_search(
    db, 'tag:new and to:PattersonPark@yahoogroups.com', '+lists',
    '+pattersonpark')
tag_search(
    db, 'tag:new and to:dev@couchdb.apache.org', '+lists', '+couchdb')
tag_search(
    db, 'tag:new and to:jira@apache.org', '+lists', '+couchdb')
tag_search(
    db, 'tag:new to:allhands@lists.canonical.com', '+lists', '+allhands')
tag_search(
    db, 'tag:new and to:remix-api@googlegroups.com', '+lists', '+echonest',
    '+musicplayers')
tag_search(
    db, 'tag:new and to:python-announce-list@python.org', '+lists', '+python',
    '+python-announce')
tag_search(db, 'tag:new and to:notmuch@notmuchmail.org', '+lists', '+notmuch',
    '-new')
tag_search(
    db, 'tag:new and to:warthogs@lists.canonical.com', '+lists', '+warthogs')
tag_search(
    db, 'tag:new and to:canonical-tech@lists.launchpad.net', '+lists',
    '+canonicaltech')
tag_search(
    db, 'tag:new and to:ubuntu-devel@lists.ubuntu.com', '+lists',
    '+ubuntudevel', '+ubuntu')
tag_search(
    db, 'tag:new and to:ubuntu-devel-discuss@lists.ubuntu.com', '+lists',
    '+ubuntudeveldiscuss', '+ubuntu')
tag_search(
    db, 'tag:new and to:mirage-list@googlegroups.com', '+lists', '+mirage')
tag_search(
    db, 'tag:new and to:python-dev@python.org', '+lists', '+python',
    '+pythondev')
tag_search(
    db, 'tag:new and to:devnull@canonical.com', '+lists', '+canonicalwiki')
tag_search(
    db, 'tag:new and to:uds-announce@lists.ubuntu.com', '+lists', '+uds')
tag_search(
    db, 'tag:new and answers.launchpad.net', '+lists', '+answers')
tag_search(
    db, 'tag:new and to:couchdb-python@googlegroups.com', '+lists', '+couchdb',
    '+python', '+couchdbpython')
tag_search(
    db, 'tag:new and to:couchdbkit@googlegroups.com', '+lists', '+couchdb',
    '+python', '+couchdbkit')
tag_search(
    db, 'tag:new and to:ubuntu-cloud@lists.ubuntu.com', '+lists',
    '+ubuntucloud')
tag_search(
    db, 'tag:new and to:tomahawk-devel@googlegroups.com', '+lists',
    '+tomahawk', '+musicplayers')
tag_search(
    db, 'tag:new and to:rhythmbox-devel@gnome.org', '+lists',
    '+rhythmbox', '+musicplayers')
tag_search(
    db, 'tag:new and to:imms-users@googlegroups.com', '+lists',
    '+imms', '+musicplayers')
tag_search(
    db, 'tag:new and to:python-dev@python.org', '+lists',
    '+python', '-new')
tag_search(
    db, 'tag:new and to:golang-nuts@googlegroups.com', '+lists',
    '+go', '-new')

#remove uninteresting notifications
tag_search(db, 'subject:"mailing list memberships reminder"', '+lists', '-new')
tag_search(db, 'tag:new and subject:"joined ubuntuone-users"', '-new')
tag_search(
    db, 'tag:new and from:noreply@launchpad.net and subject:" deactivated by"',
    '-new')
tag_search(
    db, 'tag:new and subject:"New mailing list message requiring approval for '
    'Ubunet discussion"', '-new')


# Update watch tag
for msg in q_new_msgs.search_messages():
    q = notmuch.Query(db, 'tag:unwatch and thread:%s' % msg.get_thread_id())
    for found in q.search_messages():
        logging.debug('unwatching %s', msg.get_message_id())
        msg.add_tag('unwatch')
        break

# Watched items should go to inbox
tag_search(db, 'tag:new and not tag:unwatch', '+inbox', '-new')

# Tag remaining new items for inbox
tag_search(db, 'tag:new', '-new')

end_time = time.time()
logging.info(
    'Sorted %d messages in %1.2f seconds', n_msgs, end_time - start_time)
